/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

use std::fs;
use std::iter::Iterator;
use std::os;
use std::path::Path;

use futures_lite::stream::StreamExt;
use mozjs::jsapi::{JS_DefineFunctions, JSFunctionSpec, JSObject};
use mozjs::typedarray::{CreateWith, Uint8Array};

use ion::{Context, Error, Object, Result};
use ion::flags::PropertyFlags;
use runtime::modules::NativeModule;

#[js_fn]
async unsafe fn readBinary(cx: Context, path_str: String) -> Result<*mut JSObject, ()> {
	let path = Path::new(&path_str);

	if path.is_file() {
		if let Ok(bytes) = async_fs::read(&path).await {
			rooted!(in(cx) let mut array = *Object::new(cx));
			if Uint8Array::create(cx, CreateWith::Slice(bytes.as_slice()), array.handle_mut()).is_ok()
				&& Uint8Array::create(cx, CreateWith::Length(0), array.handle_mut()).is_ok()
			{
				return Ok(array.get());
			}
		}
	}
	Err(())
}

#[js_fn]
unsafe fn readBinarySync(cx: Context, path_str: String) -> Result<*mut JSObject> {
	let path = Path::new(&path_str);

	if path.is_file() {
		if let Ok(bytes) = fs::read(&path) {
			rooted!(in(cx) let mut array = *Object::new(cx));
			if !Uint8Array::create(cx, CreateWith::Slice(bytes.as_slice()), array.handle_mut()).is_ok()
				&& !Uint8Array::create(cx, CreateWith::Length(0), array.handle_mut()).is_ok()
			{
				return Err(Error::Error(String::from("Unable to create Uint8Array")));
			}
			Ok(array.get())
		} else {
			Err(Error::Error(format!("Could not read file: {}", path_str)))
		}
	} else {
		Err(Error::Error(format!("File {} does not exist", path_str)))
	}
}

#[js_fn]
async fn readString(path_str: String) -> Result<String, ()> {
	let path = Path::new(&path_str);

	if path.is_file() {
		if let Ok(str) = async_fs::read_to_string(&path).await {
			Ok(str)
		} else {
			Err(())
		}
	} else {
		Err(())
	}
}

#[js_fn]
fn readStringSync(path_str: String) -> Result<String> {
	let path = Path::new(&path_str);

	if path.is_file() {
		if let Ok(str) = fs::read_to_string(&path) {
			Ok(str)
		} else {
			Err(Error::Error(format!("Could not read file: {}", path_str)))
		}
	} else {
		Err(Error::Error(format!("File {} does not exist", path_str)))
	}
}

#[js_fn]
async fn readDir(path_str: String) -> Result<Vec<String>, ()> {
	let path = Path::new(&path_str);

	if path.is_dir() {
		if let Ok(dir) = async_fs::read_dir(path).await {
			let entries: Vec<_> = dir.filter_map(|entry| entry.ok()).collect().await;
			let mut str_entries: Vec<String> = entries.iter().map(|entry| entry.file_name().into_string().unwrap()).collect();
			str_entries.sort();

			Ok(str_entries)
		} else {
			Ok(Vec::new())
		}
	} else {
		Err(())
	}
}

#[js_fn]
fn readDirSync(path_str: String) -> Result<Vec<String>> {
	let path = Path::new(&path_str);

	if path.is_dir() {
		if let Ok(dir) = fs::read_dir(path) {
			let entries: Vec<_> = dir.filter_map(|entry| entry.ok()).collect();
			let mut str_entries: Vec<String> = entries.iter().map(|entry| entry.file_name().into_string().unwrap()).collect();
			str_entries.sort();

			Ok(str_entries)
		} else {
			Ok(Vec::new())
		}
	} else {
		Err(Error::Error(format!("Directory {} does not exist", path_str)))
	}
}

#[js_fn]
async fn write(path_str: String, contents: String) -> Result<bool, ()> {
	let path = Path::new(&path_str);

	if !path.is_dir() {
		Ok(async_fs::write(path, contents).await.is_ok())
	} else {
		Err(())
	}
}

#[js_fn]
fn writeSync(path_str: String, contents: String) -> Result<bool> {
	let path = Path::new(&path_str);

	if !path.is_dir() {
		Ok(fs::write(path, contents).is_ok())
	} else {
		Err(Error::Error(format!("Path {} is a directory", path_str)))
	}
}

#[js_fn]
async fn createDir(path_str: String) -> Result<bool, ()> {
	let path = Path::new(&path_str);

	if !path.is_file() {
		Ok(async_fs::create_dir(path).await.is_ok())
	} else {
		Err(())
	}
}

#[js_fn]
fn createDirSync(path_str: String) -> Result<bool> {
	let path = Path::new(&path_str);

	if !path.is_file() {
		Ok(fs::create_dir(path).is_ok())
	} else {
		Err(Error::Error(format!("Path {} is a file", path_str)))
	}
}

#[js_fn]
async fn createDirRecursive(path_str: String) -> Result<bool, ()> {
	let path = Path::new(&path_str);

	if !path.is_file() {
		Ok(async_fs::create_dir_all(path).await.is_ok())
	} else {
		Err(())
	}
}

#[js_fn]
fn createDirRecursiveSync(path_str: String) -> Result<bool> {
	let path = Path::new(&path_str);

	if !path.is_file() {
		Ok(fs::create_dir_all(path).is_ok())
	} else {
		Err(Error::Error(format!("Path {} is a file", path_str)))
	}
}

#[js_fn]
async fn removeFile(path_str: String) -> Result<bool, ()> {
	let path = Path::new(&path_str);

	if path.is_file() {
		Ok(async_fs::remove_file(path).await.is_ok())
	} else {
		Err(())
	}
}

#[js_fn]
fn removeFileSync(path_str: String) -> Result<bool> {
	let path = Path::new(&path_str);

	if path.is_file() {
		Ok(fs::remove_file(path).is_ok())
	} else {
		Err(Error::Error(format!("Path {} is not a file", path_str)))
	}
}

#[js_fn]
async fn removeDir(path_str: String) -> Result<bool, ()> {
	let path = Path::new(&path_str);

	if path.is_dir() {
		Ok(fs::remove_file(path).is_ok())
	} else {
		Err(())
	}
}

#[js_fn]
fn removeDirSync(path_str: String) -> Result<bool> {
	let path = Path::new(&path_str);

	if path.is_dir() {
		Ok(fs::remove_file(path).is_ok())
	} else {
		Err(Error::Error(format!("Path {} is not a directory", path_str)))
	}
}

#[js_fn]
async fn removeDirRecursive(path_str: String) -> Result<bool, ()> {
	let path = Path::new(&path_str);

	if path.is_dir() {
		Ok(async_fs::remove_dir_all(path).await.is_ok())
	} else {
		Err(())
	}
}

#[js_fn]
fn removeDirRecursiveSync(path_str: String) -> Result<bool> {
	let path = Path::new(&path_str);

	if path.is_dir() {
		Ok(fs::remove_dir_all(path).is_ok())
	} else {
		Err(Error::Error(format!("Path {} is not a directory", path_str)))
	}
}

#[js_fn]
async fn copy(from_str: String, to_str: String) -> Result<bool, ()> {
	let from = Path::new(&from_str);
	let to = Path::new(&to_str);

	if !from.is_dir() || !to.is_dir() {
		Ok(async_fs::copy(from, to).await.is_ok())
	} else {
		Err(())
	}
}

#[js_fn]
fn copySync(from_str: String, to_str: String) -> Result<bool> {
	let from = Path::new(&from_str);
	let to = Path::new(&to_str);

	if !from.is_dir() || !to.is_dir() {
		Ok(fs::copy(from, to).is_ok())
	} else {
		Err(Error::None)
	}
}

#[js_fn]
async fn rename(from_str: String, to_str: String) -> Result<bool, ()> {
	let from = Path::new(&from_str);
	let to = Path::new(&to_str);

	if !from.is_dir() || !to.is_dir() {
		Ok(async_fs::rename(from, to).await.is_ok())
	} else {
		Err(())
	}
}

#[js_fn]
fn renameSync(from_str: String, to_str: String) -> Result<bool> {
	let from = Path::new(&from_str);
	let to = Path::new(&to_str);

	if !from.is_dir() || !to.is_dir() {
		Ok(fs::rename(from, to).is_ok())
	} else {
		Err(Error::None)
	}
}

#[js_fn]
async fn softLink(original_str: String, link_str: String) -> Result<bool, ()> {
	let original = Path::new(&original_str);
	let link = Path::new(&link_str);

	if !link.exists() {
		#[cfg(target_family = "unix")]
		{
			Ok(async_fs::unix::symlink(original, link).await.is_ok())
		}
		#[cfg(target_family = "windows")]
		{
			if original.is_file() {
				Ok(async_fs::windows::symlink_file(original, link).await.is_ok())
			} else if original.is_dir() {
				Ok(async_fs::windows::symlink_dir(original, link).await.is_ok())
			} else {
				Ok(false)
			}
		}
	} else {
		Err(())
	}
}

#[js_fn]
fn softLinkSync(original_str: String, link_str: String) -> Result<bool> {
	let original = Path::new(&original_str);
	let link = Path::new(&link_str);

	if !link.exists() {
		#[cfg(target_family = "unix")]
		{
			Ok(os::unix::fs::symlink(original, link).is_ok())
		}
		#[cfg(target_family = "windows")]
		{
			if original.is_file() {
				Ok(os::windows::fs::symlink_file(original, link).is_ok())
			} else if original.is_dir() {
				Ok(os::windows::fs::symlink_dir(original, link).is_ok())
			} else {
				Ok(false)
			}
		}
	} else {
		Err(Error::Error(String::from("Link already exists")))
	}
}

#[js_fn]
async fn hardLink(original_str: String, link_str: String) -> Result<bool, ()> {
	let original = Path::new(&original_str);
	let link = Path::new(&link_str);

	if !link.exists() {
		Ok(fs::hard_link(original, link).is_ok())
	} else {
		Err(())
	}
}

#[js_fn]
fn hardLinkSync(original_str: String, link_str: String) -> Result<bool> {
	let original = Path::new(&original_str);
	let link = Path::new(&link_str);

	if !link.exists() {
		Ok(fs::hard_link(original, link).is_ok())
	} else {
		Err(Error::Error(String::from("Link already exists")))
	}
}

const SYNC_FUNCTIONS: &[JSFunctionSpec] = &[
	function_spec!(readBinarySync, "readBinary", 1),
	function_spec!(readStringSync, "readString", 1),
	function_spec!(readDirSync, "readDir", 1),
	function_spec!(writeSync, "write", 2),
	function_spec!(createDirSync, "createDir", 1),
	function_spec!(createDirRecursiveSync, "createDirRecursive", 1),
	function_spec!(removeFileSync, "removeFile", 1),
	function_spec!(removeDirSync, "removeDir", 1),
	function_spec!(removeDirRecursiveSync, "removeDirRecursive", 1),
	function_spec!(copySync, "copy", 2),
	function_spec!(renameSync, "rename", 2),
	function_spec!(softLinkSync, "softLink", 2),
	function_spec!(hardLinkSync, "hardLink", 2),
	JSFunctionSpec::ZERO,
];

const ASYNC_FUNCTIONS: &[JSFunctionSpec] = &[
	function_spec!(readBinary, 1),
	function_spec!(readString, 1),
	function_spec!(readDir, 1),
	function_spec!(write, 2),
	function_spec!(createDir, 1),
	function_spec!(createDirRecursive, 1),
	function_spec!(removeFile, 1),
	function_spec!(removeDir, 1),
	function_spec!(removeDirRecursive, 1),
	function_spec!(copy, 2),
	function_spec!(rename, 2),
	// function_spec!(softLink, 2),
	function_spec!(hardLink, 2),
	JSFunctionSpec::ZERO,
];

#[derive(Default)]
pub struct FileSystem;

impl NativeModule for FileSystem {
	const NAME: &'static str = "fs";
	const SOURCE: &'static str = include_str!("fs.js");

	fn module(cx: Context) -> Option<Object> {
		rooted!(in(cx) let fs = *Object::new(cx));
		rooted!(in(cx) let sync = *Object::new(cx));

		if unsafe { JS_DefineFunctions(cx, fs.handle().into(), ASYNC_FUNCTIONS.as_ptr()) }
			&& unsafe { JS_DefineFunctions(cx, sync.handle().into(), SYNC_FUNCTIONS.as_ptr()) }
			&& Object::from(fs.get()).define_as(cx, "sync", sync.get(), PropertyFlags::CONSTANT_ENUMERATED)
		{
			return Some(Object::from(fs.get()));
		}
		None
	}
}
